//---------------------------------------------------------------------------
//         Algorithms used in VolgaDB Engine:
// encripting and decripting operations   (using Sapphire algorithm)
//---------------------------------------------------------------------------
//  Copyright  2000, Olga Vlasova, Russia
//  http://www.volgadb.com
//  E-mail: info@volgadb.com
//---------------------------------------------------------------------------
//  Support Kylix/Linux platform: (c) Vit (Vitaly Nevzorov), 2004, Chicago, USA
//  http://www.delphist.com
//  E-mail: vit@vingrad.ru
//---------------------------------------------------------------------------
unit VolgaAlg;

interface
uses
{$IfnDef Linux}
  Windows, Messages, 
{$EndIf non-Linux}
  SysUtils, Classes;

type
  //----- Sapphire algorithm for encripting and decripting ---------------------
  TSapphire = class(TObject)
  private
    { Private declarations }
    rotor,                            { Rotor and ratchet are used to help }
    ratchet,                          { Continually shuffle the "cards."   }
    avalanche,                        { Data dependent index. }
    last_plain,                       { Previous plain text byte. }
    last_cipher: byte;                { Previous cipher text byte. }
    cards: array[0..255] of byte;     { Array to hold a permutation of all }
    FUserKey: string;
    procedure hash_init;              { Initialize state without key. }
    procedure Burn;
    procedure SetKey(const Value: string); { Initialize state with key. }
    procedure Reoriginate;            { Restart stream. }
    function encrypt(const b: byte): byte;  { Encrypt b and return the result. }
    function decrypt(const b: byte): byte;  { Decrypt b and return the result. }
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create;
    destructor Destroy; override;
    procedure EncryptBuf(buf: PChar; size: integer);
    procedure DecryptBuf(buf: PChar; size: integer);
    property Key: string read FUserKey write SetKey;
  end;

implementation

//==============================================================================
// Encripting and decripting operations (using Sapphire algorithm)
//==============================================================================

constructor TSapphire.Create;
begin
  inherited;
  Reoriginate;
end;

destructor TSapphire.Destroy;
begin
  Burn;
  inherited;
end;

procedure TSapphire.Burn;
//destroy information about encryption key
var i: integer;
begin
  fillchar(cards, sizeof(cards), 0);
  for i := 1 to length(FUserKey) do
    FUserKey[i] := chr(0);
  FUserKey := '';
  hash_init;
end;

procedure TSapphire.hash_init;
var i, j: integer;
begin
  rotor := 1;           { Make sure we start key in a known place. }
  ratchet := 3;
  avalanche := 5;
  last_plain := 7;
  last_cipher := 11;
  j := 255;
  for i := 0 to 255 do  { Start with cards all in inverse order. }
  begin
    cards[i] := j;
    dec(j);
  end;
end;

procedure TSapphire.Reoriginate;
//initializing the cipher
{ Sets up key for use by the stream cipher, based on user_key. }
var keypos: integer;    { Index of current byte of user key. }
  toswap,               { Index for shuffling cards with. }
  swaptemp,             { For holding bytes while swapping. }
  rsum,                 { Function of key bytes processed to now. }
  i: byte;           { Indexes into key.cards. }

  function Keyrand(const limit: integer): byte;
  var u, v, mask: word;
  begin
    if limit = 0 then
      Result := 0
    else
    begin
      v := 0;
      mask := 1;
      while mask < limit do
        mask := (mask shl 1) + 1;
      repeat
        rsum := (cards[rsum] + ord(FUserKey[keypos])) and $FF;
        inc(keypos);
        if keypos > Length(FUserKey) then
        begin
          keypos := 1;
          rsum := (rsum + Length(FUserKey)) and $FF;
        end;
        u := mask and rsum;
        inc(v);
        if v > 11 then
          u := u mod limit;
      until u <= limit;
      Result := u;
    end;
  end;

begin
  if length(FUserKey) = 0 then
    hash_init
  else
  begin
    for i := 0 to 255 do  { Start with cards all in order. }
    cards[i] := i;
    keypos := 1;        { Start with first byte of user  }
//  toswap := 0;
    rsum := 0;
    for i := 255 downto 1 do  { Swap elements.  Some may get swaped back }
    begin                   { and some may get swapped with self. }
      toswap := Keyrand(i);
      swaptemp := cards[i];
      cards[i] := cards[toswap];
      cards[toswap] := swaptemp;
    end;
    rotor := cards[1];           { Make sure we start key in a known place. }
    ratchet := cards[3];
    avalanche := cards[5];
    last_plain := cards[7];
    last_cipher := cards[rsum];
//    toswap := 0;
//    swaptemp := 0;
    rsum := 0;
    keypos := 0;
  end;
end;

procedure TSapphire.SetKey(const Value: string);
//after each encrypt/decrypt operation assign new key
begin
  FUserKey := Value;
  Reoriginate;
end;

function TSapphire.Encrypt(const b: byte): byte;
var swaptemp: byte;
begin
  ratchet := (ratchet + cards[rotor]) and $FF;
  if rotor = 255 then
    rotor := 0
  else
    inc(rotor);
  swaptemp := cards[last_cipher];      { Round-robin swap. }
  cards[last_cipher] := cards[ratchet];
  cards[ratchet] := cards[last_plain];
  cards[last_plain] := cards[rotor];
  cards[rotor] := swaptemp;
  avalanche := (avalanche + cards[swaptemp]) and $FF;
  last_cipher := b xor
    cards[(cards[ratchet] + cards[rotor]) and $FF] xor
    cards[cards[(cards[last_plain] + cards[last_cipher] +
    cards[avalanche]) and $FF]];
  last_plain := b;
  encrypt := last_cipher;
end;

function TSapphire.Decrypt(const b: byte): byte;
var swaptemp: byte;
begin
  //ratchet := ratchet + cards[rotor];
  ratchet := (ratchet + cards[rotor]) and $FF;
  if rotor = 255 then
    rotor := 0
  else
    inc(rotor);
  swaptemp := cards[last_cipher];
  cards[last_cipher] := cards[ratchet];
  cards[ratchet] := cards[last_plain];
  cards[last_plain] := cards[rotor];
  cards[rotor] := swaptemp;
  avalanche := (avalanche + cards[swaptemp]) and $FF;
  //avalanche := avalanche + cards[swaptemp];
  last_plain := b xor
    cards[(cards[ratchet] + cards[rotor]) and $FF] xor
    cards[cards[(cards[last_plain] + cards[last_cipher] +
    cards[avalanche]) and $FF]];
  last_cipher := b;
  decrypt := last_plain;
end;

procedure TSapphire.EncryptBuf(buf: PChar; size: integer);
var i: integer;
begin
  for i := 0 to size - 1 do
    buf[i] := chr(encrypt(ord(buf[i])));
end;

procedure TSapphire.DecryptBuf(buf: PChar; size: integer);
var i: integer;
begin
  for i := 0 to size - 1 do
    buf[i] := chr(decrypt(ord(buf[i])));
end;

end.
