//---------------------------------------------------------------------------
//  TVolgaParser - parsing SQL text or filter expression into separate tokens
//  (identifiers, quoted strings, operands, reserved words in TList)
//  Furhter analysis is needed
//---------------------------------------------------------------------------
//  Copyright  2000, Olga Vlasova, Russia
//  http://www.volgadb.com
//  E-mail: info@volgadb.com
//---------------------------------------------------------------------------
//  Support Kylix/Lilux platform: (c) Vit (Vitaly Nevzorov), 2004, Chicago, USA
//  http://www.delphist.com
//  E-mail: vit@vingrad.ru
//---------------------------------------------------------------------------
unit VolgaLex;

interface

uses
{$IfnDef Linux}
   Windows, Messages,
{$EndIf non-Linux}
   SysUtils, Classes, DB, VolgaConst;

type
  TVolgaTokenType = (vttUnknown, vttIdent, vttQuoted, vttComma, vttPoint,
    vttReserved, vttNumber, vttAster, vttAOper, vttLOper, vttCOper, vttLBracket,
    vttRBracket, vttNull, vttCount, vttSum, vttMax, vttMin, vttAvg, vttGroup,
    vttTrue, vttFalse, vttAsterGroup);
  //Unknown identificator,Parameter,Quotated,Comma,Point,Reserved Word,Number,Asterisc
  //Arithmetic operand,Logical Operand,Comparison operand,Left bracket,Right bracket,Null
  TVolgaClause = (vcNone, vcEqual, vcGreat, vcLess, vcGreatEqual, vcLessEqual,
    vcNotEqual, vcLike, vcLeftEqual);
  TVolgaJoinType = (vjInner, vjLeft, vjRight, vjFull);

  TVolgaToken = record                  //description of the one token
    Token: string;                      //token itself
    TokenType: TVolgaTokenType;         //token type
  end;
  PVolgaToken = ^TVolgaToken;           //the token pointer

  //the class for the mask comparison
  TVolgaMask = class
  private
    FMask: Pointer;
    FSize: Integer;
  public
    constructor Create(const MaskValue: string);
    destructor Destroy; override;
    function Matches(const AText: string): Boolean;
  end;

  TVolgaCField = record                 //the structure for the storage of the information about field conditions
    Alias: string;                      //Source table alias
    Name: string;                       //name of the field
    Alias2: string;                     //second field's source table alias
    Name2: string;                      //name of the second field
    TokenType: TVolgaTokenType;         //token type
    TokenType2: TVolgaTokenType;         //token type
    Clause: TVolgaClause;               //comparison condition
    CompText: string;      //data for comparison under condition (if there's no second field)
    Maska: TVolgaMask;     //the class for the mask comparison
  end;
  PVolgaCField = ^TVolgaCField;

  TVolgaQSource = record                //the description of the source table
    Origin: string;                     //Source table
    Alias: string;                      //table's name alias
  end;
  PVolgaQSource = ^TVolgaQSource;       //table's pointer

  TVolgaQJoin = record                  //the informaion about the links between tables
    Alias: string;                      //Table1 name
    Name: string;                       //the fields of the table1 separeted by ';'
    Alias2: string;                     //Table2 name
    Name2: string;                      //the fields of the table2 separeted by ';'
    JoinType: TVolgaJoinType;           //link type
  end;
  PVolgaQJoin = ^TVolgaQJoin;

  TVolgaQColumn = record                //the description of one row of the future answer
    Alias: string;                      //Source table alias
    Name: string;                       //field name
    AsName: string;                     //field name alias
    Expr: string;                       //the expression as the source
    TokenType: TVolgaTokenType;         //token type
  end;
  PVolgaQColumn = ^TVolgaQColumn;       //row pointer

//******************************************************************************
//          TVolgaParser - syntax parser of the SQL sentences
//******************************************************************************

  TVolgaParser = class(TObject)
  private
    FJoinList: TList;
    FColumnList: TList;
    FSourceList: TList;
    FTokenList: TList;
    FWhereList: TList;
    StartToken: integer;  //StartToken
    function IsReserved(var sToken: string): Boolean;
    procedure CheckSQLFieldDef(var iAlias: string; const iName: string);
    procedure PrepareSelect;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function PrepareSQLText(AText: string): Boolean;  //returns true if Distinct token
    procedure ParseText(stxt: AnsiString);
    //prepare and return the fields descriptions in filter or WHERE request section
    function PrepareFilterExpr(FiltFields: TList; start: integer): integer;
    property SourceList: TList read FSourceList;
    property ColumnList: TList read FColumnList;
    property JoinList: TList read FJoinList;
    property TokenList: TList read FTokenList;
    property WhereList: TList read FWhereList;
  end;

//the function of the mask comparison
function LikeMask(const AText, Mask: string): Boolean;

implementation

const
  MaxCards = 30;

type
  PMaskSet = ^TMaskSet;
  TMaskSet = set of Char;
  TMaskStates = (msLiteral, msAny);
  TMaskState = record
    SkipTo: Boolean;
    case State: TMaskStates of
      msLiteral: (Literal: Char);
      msAny: ();
  end;
  PMaskStateArray = ^TMaskStateArray;
  TMaskStateArray = array[0..128] of TMaskState;

//******************************************************************************
// TVolgaMask functions
//******************************************************************************
function AnsiCharUp(t:char): char;
begin
{$IfDef Linux}
  Result:=Upcase(t);
{$Else}
  CharUpperBuff(PChar(@t),1);
  Result := t;
{$EndIf Linux}
end;

function InitMaskStates(const Mask: string;
  var MaskStates: array of TMaskState): Integer;
var
  I: Integer;
  SkipTo: Boolean;
  Literal: Char;
  P: PChar;
  Cards: Integer;

  procedure InvalidMask;
  begin
    raise Exception.CreateFmt(V_INVALIDMASK, [Mask, P - PChar(Mask) + 1]);
  end;

  procedure WriteScan(MaskState: TMaskStates);
  begin
    if I <= High(MaskStates) then
    begin
      if SkipTo then
      begin
        Inc(Cards);
        if Cards > MaxCards then InvalidMask;
      end;
      MaskStates[I].SkipTo := SkipTo;
      MaskStates[I].State := MaskState;
      if MaskState = msLiteral then MaskStates[I].Literal := AnsiCharUp(Literal);
    end;
    Inc(I);
    SkipTo := False;
  end;

begin
  P := PChar(Mask);
  I := 0;
  Cards := 0;
  SkipTo := False;
  while P^ <> #0 do
  begin
    case P^ of
      '*', '%': SkipTo := True;
      '?', '_': if not SkipTo then WriteScan(msAny);
    else
      begin
        Literal := P^;
        WriteScan(msLiteral);
      end;
    end;
    Inc(P);
  end;
  Literal := #0;
  WriteScan(msLiteral);
  Result := I;
end;

function MatchesMaskStates(const Filename: string;
  MaskStates: array of TMaskState): Boolean;
type
  TStackRec = record
    sP: PChar;
    sI: Integer;
  end;
var
  T: Integer;
  S: array[0..MaxCards - 1] of TStackRec;
  I: Integer;
  P: PChar;

  procedure Push(P: PChar; I: Integer);
  begin
    with S[T] do
    begin
      sP := P;
      sI := I;
    end;
    Inc(T);
  end;

  function Pop(var P: PChar; var I: Integer): Boolean;
  begin
    if T = 0 then
      Result := False
    else
    begin
      Dec(T);
      with S[T] do
      begin
        P := sP;
        I := sI;
      end;
      Result := True;
    end;
  end;

  function Matches(P: PChar; Start: Integer): Boolean;
  var
    I: Integer; 
  begin
    Result := False;
    for I := Start to High(MaskStates) do
      with MaskStates[I] do
      begin
        if SkipTo then
        begin
          case State of
            msLiteral:
              while (P^ <> #0) and (AnsiCharUp(P^) <> Literal) do Inc(P);
          end;
          if P^ <> #0 then Push(@P[1], I);
        end;
        case State of
          msLiteral: if AnsiCharUp(P^) <> Literal then Exit;
        end;
        Inc(P);
      end;
    Result := True;
  end;

begin
  Result := True;
  T := 0;
  P := PChar(Filename);
  I := Low(MaskStates);
  repeat
    if Matches(P, I) then Exit;
  until not Pop(P, I);
  Result := False;
end;

//******************************************************************************
// TVolgaMask
//******************************************************************************

constructor TVolgaMask.Create(const MaskValue: string);
var
  A: array[0..0] of TMaskState;
begin
  FSize := InitMaskStates(MaskValue, A);
  if FSize > 0 then begin
    FMask := AllocMem(FSize * SizeOf(TMaskState));
    InitMaskStates(MaskValue, Slice(PMaskStateArray(FMask)^, FSize));
  end;
end;

destructor TVolgaMask.Destroy;
begin
  if FMask <> nil then
    FreeMem(FMask, FSize * SizeOf(TMaskState));
end;

function TVolgaMask.Matches(const AText: string): Boolean;
begin
  Result := MatchesMaskStates(AText, Slice(PMaskStateArray(FMask)^, FSize));
end;

function LikeMask(const AText, Mask: string): Boolean;
var
  CMask: TVolgaMask;
begin
  CMask := TVolgaMask.Create(Mask);
  try
    Result := CMask.Matches(AText);
  finally
    CMask.Free;
  end;
end;

//******************************************************************************
//          TVolgaParser - SQL or Filter expression parser
// TVolgaParser separates text on tokens and checks quotations,identificators and brackets
//******************************************************************************

constructor TVolgaParser.Create;
begin
  FTokenList := TList.Create;
  FJoinList := TList.Create;
  FColumnList := TList.Create;
  FSourceList := TList.Create;
  FWhereList := TList.Create;
end;

destructor TVolgaParser.Destroy;
begin
  Clear;
  FTokenList.Free;
  FJoinList.Free;
  FColumnList.Free;
  FSourceList.Free;
  FWhereList.Free;
end;

procedure TVolgaParser.Clear;
var i: integer;
begin
  for i := 0 to FTokenList.Count - 1 do
    Dispose(PVolgaToken(FTokenList[i]));
  FTokenList.Clear;
  for i := 0 to FJoinList.Count - 1 do
    Dispose(PVolgaQJoin(FJoinList[i]));
  FJoinList.Clear;
  for i := 0 to FColumnList.Count - 1 do
    Dispose(PVolgaQColumn(FColumnList[i]));
  FColumnList.Clear;
  for i := 0 to FSourceList.Count - 1 do
    Dispose(PVolgaQSource(FSourceList[i]));
  FSourceList.Clear;
  for i := 0 to FWhereList.Count - 1 do
    Dispose(PVolgaCField(FWhereList[i]));
  FWhereList.Clear;
end;

function TVolgaParser.PrepareSQLText(AText: string): Boolean;
var i: integer;
  FF: PVolgaQColumn;
  FC: PVolgaCField;
begin
  Result := false;
  //repare tokens
  ParseText(AText);
  if FTokenList.Count < 4 then
    DatabaseError(V_FEWTOKENS);
  if PVolgaToken(FTokenList.Items[0])^.Token = 'SELECT' then
  begin
    StartToken := 1;
    if PVolgaToken(FTokenList.Items[1])^.Token = 'DISTINCT' then
      StartToken := 2;
    Result := StartToken = 2;
    //prepare lists of the answer's reply rows and tables, participating in request
    PrepareSelect;
    //check whether the list of the fields consists only from tables, that are in the tables section
    for i := 0 to FColumnList.Count - 1 do
    begin
      FF := PVolgaQColumn(FColumnList.Items[i]);
      CheckSQLFieldDef(FF^.Alias, FF^.Name);
    end;
    for i := 0 to FWhereList.Count - 1 do
    begin
      FC := PVolgaCField(FWhereList.Items[i]);
      CheckSQLFieldDef(FC^.Alias, FC^.Name);
      if FC^.TokenType2=vttIdent then  //expressions aren't checked?
        CheckSQLFieldDef(FC^.Alias2, FC^.Name2);
    end;
  end
  else
    DatabaseError(V_NOTSELECTFIRST);
end;

function TVolgaParser.IsReserved(var sToken: string): Boolean;
//is that token - the official word?
const SQLGrammar: array[0..21] of string = ('SELECT', 'FROM', 'WHERE', 'AS', 'AND',
    'OR', 'NOT', 'NULL', 'JOIN', 'LEFT', 'ON', 'LIKE', 'GROUP', 'BY',
    'SUM', 'MAX', 'MIN', 'AVG', 'TRUE', 'FALSE', 'DISTINCT', 'COUNT');
var i: integer;
begin
  Result := false;
  for i := Low(SQLGrammar) to High(SQLGrammar) do
    if UpperCase(sToken) = SQLGrammar[i] then
    begin
      Result := true;
      sToken := UpperCase(sToken);
      Exit;
    end;
end;

procedure TVolgaParser.ParseText(stxt: AnsiString);
//parse the text on the tokens and indicate token types, primary sintax analysis
var
  cCurChar, cNextChar, cQuoteChar: Char;
  sToken, FSourceT: string;
  i, iLenSQL, iTxtPos, lb, rb: Integer;
  iCurState: Integer;
  FunExpr: Boolean;   //state inwardly brackets when calling the function

const
  DefaultState = 0;
  TokenState = 1;
  QuoteState = 2;
  SpaceState = 3;
  NumberState = 4;

  procedure AddSimToSQL(cChar: Char);
  begin
    Inc(iTxtPos);
  end;

  procedure AddToken(aToken: string; typ: TVolgaTokenType);
  var pRec: PVolgaToken;
  begin
    New(pRec);                          //allocate the new structure in memory
    pRec^.Token := aToken;
    pRec^.TokenType := typ;
    FTokenList.Add(pRec);
  end;

  function ExtractFromQuotes(s: string; Quote: Char): string;
  var a, b: PChar;
  begin
    a := AllocMem(Length(s) + 1);
    a := StrPCopy(a, s);
    b := a;
    Result := AnsiExtractQuotedStr(a, Quote);
    FreeMem(b);
  end;

begin
  cQuoteChar := '''';                   //quotation symbol
  FSourceT := Trim(stxt);
  iLenSQL := Length(FSourceT);          //the length of the source SQL
  i := 1;
  lb := 0;
  rb := 0;
  FunExpr := false;                     //"inwardly argument list of the function" modifier
  iTxtPos := 1;                         //start from the 1-st position
  iCurState := DefaultState;            //current state of the viewing
  while (i <= iLenSQL) do
  begin
    cCurChar := FSourceT[i];            //current symbol
    if i = iLenSQL then
      cNextChar := #0                   //newline token
    else
      cNextChar := FSourceT[i + 1];     //following symbol
    if cCurChar = #0 then Break;
    case iCurState of
      DefaultState:
        begin
          case cCurChar of              //Char is quote
            '''', '"':
              begin
                sToken := sToken + cCurChar;
                AddSimToSQL(cCurChar);
                cQuoteChar := cCurChar;
                iCurState := QuoteState; //Current state is "Quote"
              end;
            ' ', #9, #10, #13:           //spaces or return the of the carriage
              begin
                iCurState := SpaceState;
                AddSimToSQL(' ');        //Current state is "Spaces"
              end;
            '0'..'9':
              begin
                iCurState := NumberState;
                Continue;
              end;
            '.', ',', '(', ')', '=', '>', '<', '*', '/', '+', '-': //token separator
              begin
                if FunExpr and (cCurChar<>'.')
                and (cCurChar <> ')') then            //there is invalid symbol in the argument of the function
                  DatabaseErrorFmt(V_SYMBOLINFUNC, [cCurChar]);
                AddSimToSQL(cCurChar);
                //determination of the double operations
                if ((cCurChar = '>') and (cNextChar = '=')) or
                  ((cCurChar = '<') and (cNextChar = '=')) or
                  ((cCurChar = '<') and (cNextChar = '>')) or
                  ((cCurChar = '*') and (cNextChar = '=')) then
                begin
                  AddSimToSQL(cNextChar);
                  Inc(i);               //skip double separator
                  AddToken(cCurChar + cNextChar, vttCOper); //the double comparison operation
                end
                else
                  case cCurChar of
                    '.': AddToken(cCurChar, vttPoint);
                    ',': AddToken(cCurChar, vttComma);
                    '(':
                      begin
                        AddToken(cCurChar, vttLBracket);
                        Inc(lb);
                      end;
                    ')':
                      if FunExpr then begin
                        //skip bracket and don't call AddToken
                        FunExpr := false;  //the function expression in brackets is over
                      end else
                      begin
                        AddToken(cCurChar, vttRBracket);
                        Inc(rb);
                      end;
                    '=', '>', '<': AddToken(cCurChar, vttCOper); //comparison operators
                    '*': AddToken(cCurChar, vttAster); //asterisk
                    '/', '+', '-': AddToken(cCurChar, vttAOper); //Arithmetic operators
                  end;   //!!!!!!?????? and what about negative constant?
              end;
            else
              begin
                iCurState := TokenState;
                Continue;
              end;
          end;
        end;

      //literal-number
      NumberState:
        begin
          if (cCurChar in ['0'..'9', '-', DecimalSeparator]) then
          begin
            sToken := sToken + cCurChar;
            AddSimToSQL(cCurChar);
          end
          else                          //invalid symbol in the parameter's name
            DatabaseErrorFmt(V_INVALIDNUMBER, [sToken + cCurChar]);
          if not (cNextChar in ['0'..'9', '-', DecimalSeparator]) then
          begin                         //number ended
            Inc(i);
            AddToken(sToken, vttNumber); //check the decimal separotor
            sToken := '';
            iCurState := DefaultState;
            Continue;
          end;
        end;

      //quote token
      QuoteState:
        begin                           //"Quote" state
          if (cCurChar = cQuoteChar) then //quote is closed
          begin
            if (cNextChar = cQuoteChar) then
            begin
              AddSimToSQL(cCurChar);
              Inc(i);                   //skip rabbit ears
            end
            else
            begin
              sToken := sToken + cCurChar; //add the closing quote
              sToken := ExtractFromQuotes(sToken, cQuoteChar); //extract from quotes
              AddToken(sToken, vttQuoted); //quoted text is over
              sToken := '';
              AddSimToSQL(cCurChar);
              iCurState := DefaultState;
            end;
          end
          else if cNextChar = #0 then   //have found the newline, but quote was not closed
            DatabaseError(V_QUOTENOTCLOSED)
          else
          begin
            sToken := sToken + cCurChar; //token in quote
            AddSimToSQL(cCurChar);
          end;
        end;

      //simple token
      TokenState:
        begin                           {collect the name of the token}
          if (cCurChar in ['A'..'Z', 'a'..'z', '0'..'9', '_', '$']) then
          begin
            sToken := sToken + cCurChar;
            AddSimToSQL(cCurChar);
          end
          else                          //the invalid symbol in the parameter's name
            DatabaseError(V_INVALIDIDENT);
          if not (cNextChar in ['A'..'Z', 'a'..'z', '0'..'9', '_', '$']) then
          begin                         //token is over
            Inc(i);
            if IsReserved(sToken) then
            begin //all official words are written in caps
              if (sToken = 'OR') or (sToken = 'AND') or (sToken = 'NOT') then
                AddToken(sToken, vttLOper) //logical operators
              else if sToken = 'NULL' then
                AddToken(sToken, vttNull) //word NULL
              else if sToken = 'TRUE' then
                AddToken(sToken, vttTrue) //word true
              else if sToken = 'FALSE' then
                AddToken(sToken, vttFalse) //word false
              else if cNextChar = '(' then begin
                if sToken = 'SUM' then
                  AddToken(sToken, vttSum) //function SUM
                else if sToken = 'MAX' then
                  AddToken(sToken, vttMax) //function Max
                else if sToken = 'MIN' then
                  AddToken(sToken, vttMin) //function Min
                else if sToken = 'AVG' then
                  AddToken(sToken, vttAvg) //function Avg
                else if sToken = 'COUNT' then
                  AddToken(sToken, vttCount) //function Count
                else
                  DatabaseErrorFmt(V_UNSUPPORTEDFUN, [sToken]);
                FunExpr := true;   //waiting for the closing bracket - end of the function
                Inc(i);  //skip bracket
              end
              else
                AddToken(sToken, vttReserved); //official words
            end
            else if IsValidIdent(sToken) then
              AddToken(sToken, vttIdent)
            else
              DatabaseErrorFmt(V_NOTVALIDID, [sToken]);
            sToken := '';
            iCurState := DefaultState;
            Continue;
          end;
        end;

      //spaces or return of the carriage
      SpaceState:                       //spaces and string's separators are skiped
        if not (cCurChar in [' ', #9, #10, #13]) then
        begin                           //spaces are over
          iCurState := DefaultState;
          Continue;
        end;
    end;
    Inc(i);                             //go to the next symbol
  end;
  if lb <> rb then DatabaseError(V_ERBRACKET);
end;

function TVolgaParser.PrepareFilterExpr(FiltFields: TList; start: integer): integer;
//making pairs from tokens for comparison, return - the number of the token: GROUP or 0
var iToc: integer;
  CurToc, NextToc: PVolgaToken;
  tbl, fld, tbl2, fld2: string;
  vc: TVolgaClause;
  tok,tok2: TVolgaTokenType;
  iState: integer;
  negative: boolean;
const
  TableState = 0;
  FieldState = 1;                       //field's names follow
  OperState = 2;                        //operand follows
  ExprState = 3;                        //expression
  Field2State = 4;                      //the field's name follows
  LogopState = 5;                       //the logical connective between pairs

  procedure AddPair;
  var C: PVolgaCField;
  begin                                 //add pair to list
    New(C);
    C^.Alias := tbl;
    C^.Name := fld;
    C^.Alias2 := tbl2; //if table's alias does not comply  with, it is JOIN!!!!!!!!!!!!!!
    C^.Name2 := fld2;
    C^.Clause := vc;
    C^.CompText := '';
    C^.TokenType := tok;
    C^.TokenType2 := tok2;
    C^.Maska := nil;
    FiltFields.Add(C);
  end;

  procedure ClearTemp;
  begin
    fld := '';
    tbl := '';
    vc := vcNone;
    tbl2 := '';
    fld2 := '';
    tok := vttUnknown;
    tok2 := vttUnknown;
  end;

  procedure ErrorExpr;
  begin
    DatabaseError(V_ERRCOMPARING);
  end;

begin
  iToc := start;
  iState := TableState;
  Result := 0;
  negative := false;
  ClearTemp;
  while iToc < TokenList.Count do
  begin
    CurToc := PVolgaToken(TokenList.Items[iToc]);
    if iToc = FTokenList.Count - 1 then
      NextToc := nil                    //the newline token
    else
      NextToc := PVolgaToken(FTokenList.Items[iToc + 1]); //next symbol
    //comparison condition is: [table.]<field>operator<value> AND [table.]<field>operator<value>....
    case iState of
      TableState:
        begin
          if (CurToc^.TokenType = vttIdent) and (NextToc^.TokenType = vttPoint) then
          begin
            tbl := CurToc^.Token;       //it is table's name
            inc(iToc, 2);               //go to field's name
          end;
          iState := FieldState;         //then comes field's name
          Continue;
        end;
      FieldState:                       //field's name
        case CurToc^.TokenType of
          vttIdent, vttQuoted:
            begin
              fld := CurToc^.Token;
              iState := OperState;
              tok := CurToc^.TokenType;
              inc(iToc);
            end;
          else
            ErrorExpr;                  //error
        end;
      Field2State:                      //field's name or comparison expression
        case CurToc^.TokenType of
          vttIdent, vttQuoted, vttNumber, vttNull, vttTrue, vttFalse:
            begin
              fld2 := CurToc^.Token;
              if negative then fld2 := '-' + fld2;
              negative := false; 
              iState := LogopState;
              Inc(iToc);
              tok2 := CurToc^.TokenType;
              AddPair;                  //add the pair in pair list
              ClearTemp;
            end;
          else
            ErrorExpr;
        end;
      OperState:                        //operand
        case CurToc^.TokenType of
          vttReserved:
            if CurToc^.Token = 'LIKE' then
            begin
              vc := vcLike;
              iState := ExprState;
              Inc(iToc);
            end;
          vttCOper:                     //comparison operand
            begin
              if CurToc^.Token = '=' then
                vc := vcEqual
              else if CurToc^.Token = '>' then
                vc := vcGreat
              else if CurToc^.Token = '<' then
                vc := vcLess
              else if CurToc^.Token = '>=' then
                vc := vcGreatEqual
              else if CurToc^.Token = '<=' then
                vc := vcLessEqual
              else if CurToc^.Token = '<>' then
                vc := vcNotEqual
              else if CurToc^.Token = '*=' then
                vc := vcLeftEqual
              else
                ErrorExpr;
              iState := ExprState;
              Inc(iToc);
            end;
          else
            ErrorExpr;                  //the rest is error
        end;
      ExprState:                        //expression for comparison
        case CurToc^.TokenType of
          vttAOper:
            if (CurToc^.Token = '-') and (NextToc^.TokenType = vttNumber) then
            begin
              negative := true;
              Inc(iToc);
              iState := Field2State;
              Continue;
            end;
          vttIdent:
            begin
              if (NextToc <> nil) and (NextToc^.TokenType = vttPoint) then
              begin
                tbl2 := CurToc^.Token;  //it is table's name
                inc(iToc, 2);           //go to field's name
              end;
              iState := Field2State;
              Continue;
            end;
          else
            iState := Field2State;
        end;
      LogopState:                       //logical operand - a connective between pairs
        if CurToc^.Token = 'AND' then
        begin
          iState := TableState;
          Inc(iToc);
        end
        else if (CurToc^.Token = 'GROUP') and (NextToc <> nil) and (NextToc^.Token = 'BY') then
        begin
          Result := iToc + 1;  //number of the token, that follows after the word GROUP
          Break;  //begining of the GROUP BY section, WHERE part is over
        end
        else
          ErrorExpr;
    end;
  end;
end;

procedure TVolgaParser.PrepareSelect;
//parsing of the request on parts

var iToc: integer;
  CurToc, NextToc: PVolgaToken;
  iColumn: PVolgaQColumn;
  iTable: PVolgaQSource;
//iJoin: PVolgaQJoin;
  IsColumn, IsTable, {IsJoin,} IsAlias: Boolean;
  iState: integer;
  grptable,grpfield:string;
label Zv;
const
  TableState = 0;                      //table's name ends with the point
  ColumnState = 1;                     //field's rows of the future answer are coming
  FromState = 2;                       //tables are listing
  JoinState = 3;                       //links are listing
  WhereState = 4;                      //conditions are listing
  FuncTable = 5;                       //waiting for the table's name in brankets after function
  FuncColumn = 6;                      //waiting for the field's name in brankets after function
  GroupState = 7;                      //GROUP BY section

  procedure AddNewTable;
  begin
    New(iTable);
    iTable^.Origin := '';
    iTable^.Alias := '';
  end;

  procedure AddNewColumn;
  begin
    New(iColumn);
    iColumn^.AsName := '';
    iColumn^.Name := '';
    iColumn^.Alias := '';
    iColumn^.Expr := '';
    iColumn^.TokenType := vttUnknown;
  end;

  procedure AddGroupField;
  //one row's description contains message that this field serves for the grouping
  var k: integer;
    Col: PVolgaQColumn;
    IsFound: Boolean;
  begin
    IsFound := false;
    for k := 0 to FColumnList.Count - 1 do begin
      Col := PVolgaQColumn(FColumnList.Items[k]);
      if (Col^.AsName = grpfield) and ((grptable = '') or (grptable = Col^.Alias)) then
        if Col^.TokenType in [vttCount, vttSum, vttMax, vttMin, vttAvg, vttGroup] then
          DatabaseErrorFmt(V_ERRGROUPFIELD, [grpfield])
        else begin
          Col^.TokenType := vttGroup;
          IsFound := true;
          Break;
        end;
    end;
    if not IsFound then
      DatabaseErrorFmt(V_ERRGROUPFIELD, [grpfield]);
    grptable := '';
    grpfield := '';
  end;

begin
  iToc := StartToken;
  iState := TableState;
  IsColumn := false;
  IsTable := false;
//IsJoin := false;
  IsAlias := false;
  grptable := '';
  grpfield := '';
  //first after the official word can be: *, table name, number, function's name or parameter
  //after identificator can be: point, comma, bracket (it is function), arithmetic operator or AS
  while iToc < FTokenList.Count do
  begin
    CurToc := PVolgaToken(FTokenList.Items[iToc]);
    if iToc = FTokenList.Count - 1 then
      NextToc := nil                    //newline token
    else
      NextToc := PVolgaToken(FTokenList.Items[iToc + 1]); //next symbol
    case iState of
      TableState, FuncTable:
        begin
          if (CurToc^.TokenType = vttIdent) and (NextToc^.TokenType = vttPoint) then
          begin
            if not IsColumn then AddNewColumn;
            IsColumn := true;
            iColumn^.Alias := CurToc^.Token; //it is table's name
            inc(iToc, 2);               //go to field's name
          end;
          if iState = TableState then
            iState := ColumnState         //then comes field's name
          else
            iState := FuncColumn;
          Continue;
        end;
      //fields description section
      ColumnState, FuncColumn:
        begin
          case CurToc^.TokenType of
            vttIdent, vttQuoted, vttNumber, vttAster: //next token is important
              case NextToc^.TokenType of
                vttComma:               //go to the next row
                  begin
                    if not IsColumn then AddNewColumn;
                    if not IsAlias then begin//there was not AS before it
                      iColumn^.Name := CurToc^.Token;
                      if iColumn^.TokenType = vttUnknown then
                        iColumn^.TokenType := CurToc^.TokenType;
                    end;
                    iColumn^.AsName := CurToc^.Token; //alias is always there!
                    FColumnList.Add(iColumn);
                    Inc(iToc, 2);       //skip comma
                    IsColumn := false;  //the row is over
                    IsAlias := false;
                    iState := TableState;         //then comes table/field name
                    Continue;
                  end;
//              vttRBracket: ;         //it is closing bracket in function
                vttAOper: ;            //it is arithmetic expression
                vttReserved:           //is it AS or FROM?
                  begin
                    if NextToc^.Token = 'AS' then
                    begin
                      if not IsColumn then AddNewColumn;
                      IsColumn := true;
                      iColumn^.Name := CurToc^.Token;
                      if iColumn^.TokenType = vttUnknown then
                        iColumn^.TokenType := CurToc^.TokenType;
                      IsAlias := true;
                      Inc(iToc, 2);     //skip AS
                      Continue;
                    end
                    else if NextToc^.Token = 'FROM' then
                    begin
                      if not IsColumn then AddNewColumn;
                      if not IsAlias then begin//there was not AS before it
                        iColumn^.Name := CurToc^.Token;
                        if iColumn^.TokenType = vttUnknown then
                          iColumn^.TokenType := CurToc^.TokenType;
                      end;
                      iColumn^.AsName := CurToc^.Token;
                      FColumnList.Add(iColumn);
                      iState := FromState;
                      Inc(iToc, 2);     //skip FROM
                      IsColumn := false;
                      IsAlias := false;
                      Continue;
                    end
                    else
                      DatabaseErrorFmt(V_NEXTTOKENINVALID,[CurToc^.Token, NextToc^.Token]);
                  end;
                else
                  DatabaseErrorFmt(V_NEXTTOKENINVALID,[CurToc^.Token, NextToc^.Token]);
              end;
            vttCount, vttSum, vttMax, vttMin, vttAvg:  //static function
              begin
                AddNewColumn;
                iColumn^.TokenType := CurToc^.TokenType;  //function type
                IsColumn := true;
                iState := FuncTable;         //then comes table/field name
                Inc(iToc);
              end;
  //        vttNumber: Inc(iToc);  //may be this is something like: 0.0 as Name
            vttReserved:
              DatabaseErrorFmt(V_ERRINSELECTPART,[CurToc^.Token]);
            vttLBracket: Inc(iToc);
            vttNull: Inc(iToc);
            else
              DatabaseErrorFmt(V_ERRINSELECTPART,[CurToc^.Token]);
          end;                          //case CurToc^.TokenType of
        end;                            //ColumnState:

      //tables description section
      FromState:
        begin
          case CurToc^.TokenType of
            vttIdent, vttQuoted: //next token is important
              if NextToc <> nil then
                case NextToc^.TokenType of
                  vttComma:
                    begin
                      if not IsTable then AddNewTable;
                      if not IsAlias then //there was no token before it
                        iTable^.Origin := CurToc^.Token;
                      iTable^.Alias := CurToc^.Token; //alias is always there!
                      FSourceList.Add(iTable);
                      Inc(iToc, 2);     //skip comma
                      IsTable := false; //the table is over
                      IsAlias := false;
                      Continue;
                    end;
                  vttReserved:          //is it WHERE or JOIN?
                    if (NextToc^.Token = 'WHERE') or (NextToc^.Token = 'GROUP') then
                    begin
                      if not IsTable then AddNewTable;
                      if not IsAlias then //there was not token before it
                        iTable^.Origin := CurToc^.Token;
                      iTable^.Alias := CurToc^.Token; //alias is always there!
                      FSourceList.Add(iTable);
                      Inc(iToc, 2);     //skip Where or Group
                      IsTable := false; //the table is over
                      IsAlias := false;
                      if (NextToc^.Token = 'WHERE') then
                        iState := WhereState
                      else
                        iState := GroupState;
                      Continue;
                    end
                    else if NextToc^.Token = 'JOIN' then
                    begin
                      iState := JoinState;
                    end
                    else
                      ;
                  vttIdent, vttQuoted:  //then comes the alias of the table's name
                    begin
                      if IsTable then
                        DatabaseErrorFmt(V_ERRINFROMPART,[CurToc^.Token]);
                      AddNewTable;
                      iTable^.Origin := CurToc^.Token;
                      Inc(iToc);
                      IsTable := true;
                      IsAlias := true;
                      Continue;
                    end;
                  else
                    DatabaseErrorFmt(V_ERRINFROMPART,[CurToc^.Token]);
                end
              else                      //NextToc = nil
              begin
                if not IsTable then AddNewTable;
                if not IsAlias then     //there was no token before it
                  iTable^.Origin := CurToc^.Token;
                iTable^.Alias := CurToc^.Token;
                FSourceList.Add(iTable);
                Break;
              end;
            else
              DatabaseErrorFmt(V_ERRINFROMPART,[CurToc^.Token]);
          end;
        end;
      JoinState: break;                 //links are listing
      WhereState:                       //conditions are listing
        begin
          iToc := PrepareFilterExpr(FWhereList, iToc);
          if iToc = 0 then
            break                        //is there something after WHERE?
          else if iToc < FTokenList.Count - 1 then
           iState := GroupState        //pointer after word GROUP
          else
            DatabaseError(V_ERRINGROUPPART);
        end;
      GroupState:                       //grouping section for stastic functions
        begin                           //here fields must be listed as in SELECT section
          case CurToc^.TokenType of
            vttReserved:
              if CurToc^.Token <> 'BY' then
                DatabaseError(V_ERRINGROUPPART)
              else if grptable + grpfield <> '' then
           //     Inc(iToc)  //if there were no fields yet
           //   else
                DatabaseError(V_ERRINGROUPPART);
            vttIdent, vttQuoted:
              if NextToc <> nil then
                case NextToc^.TokenType of
                  vttComma:  //it is field's name
                    begin
                      grpfield := CurToc^.Token;
                      AddGroupField;
                      Inc(iToc);  //skip comma
                    end;
                  vttPoint:  //it is  table's name
                    begin
                      grptable := CurToc^.Token;
                      Inc(iToc);  //skip point
                    end;
                  else
                    DatabaseError(V_ERRINGROUPPART)
                end
              else  //the sentence is over
              begin
                grpfield := CurToc^.Token;
                AddGroupField;
                Break;
              end;
            else
              DatabaseError(V_ERRINGROUPPART);
          end;
          Inc(iToc);
        end;
    end;                                //case iState of
  end;                                  //while iToc < FTokenList.Count

  if (FSourceList.Count = 0) or (FColumnList.Count = 0) then
    DatabaseError(V_ERRNOALLPARTS);
end;

//check whether fields correspondes tables

procedure TVolgaParser.CheckSQLFieldDef(var iAlias: string; const iName: string);
var OneAlias: string;
  j: integer;
  IsFound: Boolean;
begin
  OneAlias := PVolgaQSource(FSourceList[0])^.Alias; //the 1-st table from the tables list
  IsFound := true;
  if iAlias = '' then
    iAlias := OneAlias
  else
  begin            //alias is specified earlier - check the presence of that table
    for j := 0 to FSourceList.Count - 1 do
    begin
      IsFound := AnsiCompareText(iAlias,PVolgaQSource(FSourceList[j])^.Alias)=0;
      if IsFound then Break;
    end;
    if not IsFound then
      DatabaseErrorFmt(V_ERRFIELDNAME,[iAlias,iName]);
  end;
end;

end.

